define(function () { 'use strict';

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */

    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }

    function __generator(thisArg, body) {
        var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
        return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
        function verb(n) { return function (v) { return step([n, v]); }; }
        function step(op) {
            if (f) throw new TypeError("Generator is already executing.");
            while (_) try {
                if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
                if (y = 0, t) op = [op[0] & 2, t.value];
                switch (op[0]) {
                    case 0: case 1: t = op; break;
                    case 4: _.label++; return { value: op[1], done: false };
                    case 5: _.label++; y = op[1]; op = [0]; continue;
                    case 7: op = _.ops.pop(); _.trys.pop(); continue;
                    default:
                        if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                        if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                        if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                        if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                        if (t[2]) _.ops.pop();
                        _.trys.pop(); continue;
                }
                op = body.call(thisArg, _);
            } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
            if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
        }
    }

    /**
     * Formats an enum value as a string for debugging and debug assertions.
     */
    function formatEnum(ts, value, enumObject, isFlags) {
        if (value === void 0) { value = 0; }
        var members = getEnumMembers(enumObject);
        if (value === 0) {
            return members.length > 0 && members[0][0] === 0 ? members[0][1] : "0";
        }
        if (isFlags) {
            var result = "";
            var remainingFlags = value;
            for (var i = members.length - 1; i >= 0 && remainingFlags !== 0; i--) {
                var _a = members[i], enumValue = _a[0], enumName = _a[1];
                if (enumValue !== 0 && (remainingFlags & enumValue) === enumValue) {
                    remainingFlags &= ~enumValue;
                    result = "" + enumName + (result ? "|" : "") + result;
                }
            }
            if (remainingFlags === 0) {
                return result;
            }
        }
        else {
            for (var _i = 0, members_1 = members; _i < members_1.length; _i++) {
                var _b = members_1[_i], enumValue = _b[0], enumName = _b[1];
                if (enumValue === value) {
                    return enumName;
                }
            }
        }
        return value.toString();
        function getEnumMembers(enumObject) {
            var result = [];
            for (var name_1 in enumObject) {
                var value_1 = enumObject[name_1];
                if (typeof value_1 === "number") {
                    result.push([value_1, name_1]);
                }
            }
            return stableSort(result, function (x, y) { return compareValues(x[0], y[0]); });
        }
        function compareComparableValues(a, b) {
            return a === b
                ? 0 /* EqualTo */
                : a === undefined
                    ? -1 /* LessThan */
                    : b === undefined
                        ? 1 /* GreaterThan */
                        : a < b
                            ? -1 /* LessThan */
                            : 1 /* GreaterThan */;
        }
        function compareValues(a, b) {
            return compareComparableValues(a, b);
        }
        function stableSortIndices(array, indices, comparer) {
            // sort indices by value then position
            indices.sort(function (x, y) { return comparer(array[x], array[y]) || compareValues(x, y); });
        }
        function stableSort(array, comparer) {
            var indices = array.map(function (_, i) { return i; });
            stableSortIndices(array, indices, comparer);
            return indices.map(function (i) { return array[i]; });
        }
    }

    var makePlugin = function (utils) {
        var scanContainer = undefined;
        var customPlugin = {
            id: "scanner",
            displayName: "Scanner",
            didMount: function (sandbox, container) {
                // Create a design system object to handle
                // making DOM elements which fit the playground (and handle mobile/light/dark etc)
                var ds = utils.createDesignSystem(container);
                ds.title("Scanner Results");
                ds.p("This plugin will run the TypeScript scanner over a code in your editor.");
                ds.p("This isn't the same as what TypeScript is actually doing (because the Parser would handle re-scanning) but it's enough to have a sense of what is going on under the hood.");
                scanContainer = document.createElement("div");
                var scanDS = utils.createDesignSystem(scanContainer);
                scanDS.p("Change the text in your editor to see the results.");
                container.appendChild(scanContainer);
            },
            modelChangedDebounce: function (sandbox, model) { return __awaiter(void 0, void 0, void 0, function () {
                function isJSDocLikeText(text, start) {
                    return text.charCodeAt(start + 1) === 42 && text.charCodeAt(start + 2) === 42 && text.charAt(start + 3) !== "/";
                }
                var scanDS, ts, isJSX, scanner, tokens, kind, start, _loop_1, renderTokens;
                return __generator(this, function (_a) {
                    if (!scanContainer)
                        return [2 /*return*/];
                    scanDS = utils.createDesignSystem(scanContainer);
                    ts = sandbox.ts;
                    isJSX = sandbox.filepath.endsWith("x") ? 1 : 0;
                    scanner = ts.createScanner(sandbox.getCompilerOptions().target, false, isJSX);
                    scanner.setText(model.getValue());
                    scanner.setOnError(function (message, length) {
                        scanDS.clear();
                        scanDS.subtitle("Scan failed");
                        scanDS.p(message);
                    });
                    tokens = [];
                    start = 0;
                    _loop_1 = function () {
                        kind = scanner.scan();
                        var end = scanner.getTextPos();
                        var token = {
                            kind: kind,
                            token: formatEnum(ts, kind, ts.SyntaxKind),
                            start: start,
                            end: end,
                        };
                        var jsdocTokens = [];
                        // Handle JSDoc parsing, basically the same thing as above, but uses scanJsDocToken
                        if (kind === ts.SyntaxKind.MultiLineCommentTrivia && isJSDocLikeText(scanner.getText(), start)) {
                            scanner.scanRange(start + 3, end - start - 5, function () {
                                var token = scanner.scanJsDocToken();
                                var jsDocStart = start + 3;
                                while (token != ts.SyntaxKind.EndOfFileToken) {
                                    var jsDocEnd = scanner.getTextPos();
                                    jsdocTokens.push({
                                        kind: kind,
                                        token: formatEnum(ts, token, ts.SyntaxKind),
                                        start: jsDocStart,
                                        end: jsDocEnd,
                                    });
                                    jsDocStart = jsDocEnd;
                                    token = scanner.scanJsDocToken();
                                }
                            });
                        }
                        if (jsdocTokens.length)
                            token.jsdocTokens = jsdocTokens;
                        tokens.push(token);
                        start = end;
                    };
                    while (kind != ts.SyntaxKind.EndOfFileToken) {
                        _loop_1();
                    }
                    scanDS.clear();
                    scanDS.subtitle("Scan results");
                    renderTokens = function (container, tokens) {
                        var ul = document.createElement("ul");
                        container.appendChild(ul);
                        tokens.forEach(function (t, i) {
                            var li = document.createElement("li");
                            ul.appendChild(li);
                            if (t.jsdocTokens) {
                                li.innerHTML = "<code>" + t.token + "</code><p>JSDoc:</p>";
                                renderTokens(li, t.jsdocTokens);
                            }
                            else {
                                li.innerHTML = "<code>" + t.token + "</code>";
                                li.onmouseover = function () {
                                    var beforeIndex = i - 1;
                                    var startPos = model.getPositionAt(0);
                                    if (beforeIndex >= 0) {
                                        var beforeToken = tokens[beforeIndex];
                                        startPos = model.getPositionAt(beforeToken.end);
                                    }
                                    var endPos = model.getPositionAt(t.end);
                                    sandbox.editor.setSelection({
                                        selectionStartLineNumber: startPos.lineNumber,
                                        selectionStartColumn: startPos.column,
                                        positionLineNumber: endPos.lineNumber,
                                        positionColumn: endPos.column,
                                    });
                                };
                            }
                        });
                    };
                    renderTokens(scanContainer, tokens);
                    return [2 /*return*/];
                });
            }); },
        };
        return customPlugin;
    };

    return makePlugin;

});
